__local_schema = r"""<?xml version="1.0" encoding="UTF-8"?>
<!-- edited with XMLSpy v2019 (http://www.altova.com) by Michael Riffle (University of Washington) -->
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:vc="http://www.w3.org/2007/XMLSchema-versioning" elementFormDefault="qualified" attributeFormDefault="unqualified" version="1.0" vc:minVersion="1.1">
	<xs:element name="proxl_input">
		<xs:annotation>
			<xs:documentation>Root element for all data</xs:documentation>
		</xs:annotation>
		<!-- Element name must match string in Java class XSD_ElementAttributeNamesConstants  -->
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_program_info"/>
				<xs:element ref="linkers">
					<xs:annotation>
						<xs:documentation>The chemical linkers used in this experiment. There must be at least one.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="reported_peptides">
					<xs:annotation>
						<xs:documentation>All the reported peptides from a PSM search. A reported peptide is a string that uniquely identifies an identified peptide (or linked pair of peptides) IN THIS SEARCH. (This string is not used by ProXL for this purpose.) A unique peptide i this context is one or two peptide sequences linked together at specific positions in those sequences, and with modifications at specific sites. E.g. PEP[16.01]TIDE would uniquely ID a peptide with this sequence with the given mass shift on the given position. or KEKTIDE(3)--PEPT[16.1]IDE(2) would uniquely identify the crosslink between two peptides with these ssequences at positions 3 and 2 (respectively) and where the second peptide has the shown modification. There is no syntax expected for reported peptide, but the reported peptide string must be unique, and no two reported peptides must describe the same peptide or pair of peptides linked at the same positions with the same modifications.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="matched_proteins" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Contains the protein sequences matched by peptides found in the experiment, and associated names, descriptions, and taxonomy IDs for those proteins. This will typically come from the FASTA file used to search the data. Only the proteins matched by the data should be included, not all the proteins in the FASTA file.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="static_modifications" minOccurs="0"/>
				<xs:element ref="decoy_labels" minOccurs="0">
					<xs:annotation>
						<xs:documentation>An optional list of decoy labels (strings that are prefixes on protein names that indicate the given protein is a decoy (e.g., "DECOY_") </xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="configuration_files" minOccurs="0">
					<xs:annotation>
						<xs:documentation>(Optional) The configuration files associated with the search.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
			<xs:attribute name="fasta_filename" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The filename (not including the path) of the FASTA file used to search the data.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="name" type="xs:string">
				<xs:annotation>
					<xs:documentation>An optional name to be used for this search in ProXL. If not present, a default will be used. This name may be changed within ProXL.</xs:documentation>
				</xs:annotation>
				<!-- Attribute name must match string in Java class XSD_ElementAttributeNamesConstants  -->
			</xs:attribute>
			<xs:attribute name="comment" type="xs:string">
				<xs:annotation>
					<xs:documentation>An optional comment to be associated with this search when it is imported into ProXL.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="search_program">
		<xs:annotation>
			<xs:documentation>A search program, such as Kojak or XQuest</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element name="psm_annotation_types">
					<xs:annotation>
						<xs:documentation>The types of annotations this search program has for PSMs--such as quality scores.</xs:documentation>
					</xs:annotation>
					<xs:complexType>
						<xs:sequence>
							<xs:element ref="filterable_psm_annotation_types"/>
							<xs:element ref="descriptive_psm_annotation_types" minOccurs="0"/>
						</xs:sequence>
					</xs:complexType>
				</xs:element>
				<xs:element name="psm_per_peptide_annotation_types" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Optional. If this scoring algorithm reports a separate score for each of two cross-linked peptides in a PSM, this element should include the types of scores/annotations to be associated with each peptide. Note that if per-peptide annotations are reported for cross-linked peptides, it is expected that the same annotation is present for non cross-linked peptides as well.</xs:documentation>
					</xs:annotation>
					<xs:complexType>
						<xs:sequence>
							<xs:element ref="filterable_psm_per_peptide_annotation_types">
								<xs:annotation>
									<xs:documentation>If the search program supports it, the collection of numeric annotation types (such as scores) that may be associated with a each separate peptide that is part that are part of a single PSM (e.g. two cross-linked PSMs)</xs:documentation>
								</xs:annotation>
							</xs:element>
							<xs:element ref="descriptive_psm_per_peptide_annotation_types" minOccurs="0">
								<xs:annotation>
									<xs:documentation>Optional. The collection of annotation types associated with each separate peptide that is part of a single PSM that are descriptive or not meant to be used to filter based on quality</xs:documentation>
								</xs:annotation>
							</xs:element>
						</xs:sequence>
					</xs:complexType>
				</xs:element>
				<xs:element name="reported_peptide_annotation_types" minOccurs="0">
					<xs:annotation>
						<xs:documentation>If the search program reports a peptide-level score for each distinct peptide (e.g., peptide-level scores in Percolator), these are the types of scores/annotations that are reported at that level.</xs:documentation>
					</xs:annotation>
					<xs:complexType>
						<xs:sequence>
							<xs:element ref="filterable_peptide_annotation_types" minOccurs="0"/>
							<xs:element ref="descriptive_peptide_annotation_types" minOccurs="0"/>
						</xs:sequence>
					</xs:complexType>
				</xs:element>
			</xs:sequence>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the search program, e.g., "Kojak" or "XQuest" Used to reference this search program in this document.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="display_name">
				<xs:annotation>
					<xs:documentation>If present, the value for this attribute will be used to display the name of this program in the web site. If abscent the value for "name" will be used.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="version" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The version of the above program that was run to generate these results.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>(Optional) A description of the search program.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="reported_peptides">
		<xs:annotation>
			<xs:documentation>All the reported peptides from a PSM search</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="reported_peptide" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>A reported peptide. See the annotation for "reported_peptides" for a detailed explanation.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="reported_peptide">
		<xs:annotation>
			<xs:documentation>A peptide from a peptide spectrum match, may be unlinked, looplinked, or a crosslinked pair of peptides.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="peptides">
					<xs:annotation>
						<xs:documentation>A list of peptides associated with a reported peptide in a PSM search. For example, a reported peptide that is a crosslinked pair of peptides will have two peptides associated with it. A looplink or unlink will have one.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="psms">
					<xs:annotation>
						<xs:documentation>The PSMs found for this reported peptide in this search.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element name="reported_peptide_annotations" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Any annotations generated at the reported-peptide-level for this reported peptide.</xs:documentation>
					</xs:annotation>
					<xs:complexType>
						<xs:sequence>
							<xs:element ref="filterable_reported_peptide_annotations" minOccurs="0"/>
							<xs:element ref="descriptive_reported_peptide_annotations" minOccurs="0"/>
						</xs:sequence>
					</xs:complexType>
				</xs:element>
			</xs:sequence>
			<xs:attribute name="reported_peptide_string" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The string that uniquely identifies this peptide (or peptides) linked at these specific positions (if any), with the given positions and respective masses of modifications (if any). See the annotation of "reported_peptides" for more details.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="type" type="link_type" use="required">
				<xs:annotation>
					<xs:documentation>The type of reported peptide, either "crosslink" for two peptides linked by a crosslinker, "looplink" for a single peptide connected by two ends of a crosslinker, or "unlinked" for peptides that are neither of these.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_reported_peptide_annotations">
		<xs:annotation>
			<xs:documentation>The numeric annotations (usually scores of some type) assigned to a reported peptide in the experiment that are meant to describe the likelihood of the peptide being present given all PSMs. Intended to be used to filter out poor matches. Examples include p-value, q-value, PEP</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="filterable_reported_peptide_annotation" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>An annotation or score assigned to a reported peptide</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_reported_peptide_annotations">
		<xs:annotation>
			<xs:documentation>The descriptive annotations (string or numeric) assigned to a reported peptide in the experiment. Not intended to be used to filter out poor matches.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="descriptive_reported_peptide_annotation" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>An annotation or score assigned to a reported peptide</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_reported_peptide_annotation">
		<xs:annotation>
			<xs:documentation>A numeric annotation (usually a score of some type) assigned to a reported peptide in the experiment that are meant to describe the likelihood of the peptide being present given all PSMs. Intended to be used to filter out poor matches. Examples include p-value, q-value, PEP</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the search program, must be found in the search_programs element</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the peptide annotation type (e.g., "q-value"). Must be found in the possible score types for the above search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="value" type="xs:decimal" use="required">
				<xs:annotation>
					<xs:documentation>The value or score assigned for the given annotation type from the given search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_reported_peptide_annotation">
		<xs:annotation>
			<xs:documentation>A descriptive annotations (string or numeric) assigned to a reported peptide in the experiment. Not intended to be used to filter out poor matches.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the search program, must be found in the search_programs element</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the peptide annotation type (e.g., "q-value"). Must be found in the possible score types for the above search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="value" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The value or score assigned for the given annotation type from the given search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="peptides">
		<xs:annotation>
			<xs:documentation>A list of peptides associated with a reported peptide in a PSM search.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="peptide" maxOccurs="2">
					<xs:annotation>
						<xs:documentation>A peptide parsed from the reported peptide. Unlinked and looplinked reported peptides will result in one peptide. Crosslinked reported peptides will result in two peptides.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="peptide">
		<xs:annotation>
			<xs:documentation>A peptide parsed from the reported peptide.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="modifications" minOccurs="0"/>
				<xs:element ref="linked_positions" minOccurs="0">
					<xs:annotation>
						<xs:documentation>The list of linked positions in a peptide. Do not include monolinks here, they are listed as modifications in the modifications element. Only list looplinked and crosslinked positions.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element name="peptide_isotope_labels" minOccurs="0">
					<xs:annotation>
						<xs:documentation>If this peptide is isotopically labelled, this element contains the label present on the peptide.</xs:documentation>
					</xs:annotation>
					<xs:complexType>
						<xs:sequence>
							<xs:element name="peptide_isotope_label" minOccurs="0">
								<xs:annotation>
									<xs:documentation>The isotope label for for this peptide. </xs:documentation>
								</xs:annotation>
								<xs:complexType>
									<xs:attribute name="label" use="required">
										<xs:annotation>
											<xs:documentation>The isotope label. Should be one of: 13C, 15N, 18O, or 2H</xs:documentation>
										</xs:annotation>
									</xs:attribute>
								</xs:complexType>
							</xs:element>
						</xs:sequence>
					</xs:complexType>
				</xs:element>
			</xs:sequence>
			<xs:attribute name="sequence" type="sequence_type" use="required">
				<xs:annotation>
					<xs:documentation>The sequence of the peptide, without any special characters.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="unique_id">
				<xs:annotation>
					<xs:documentation>Optional. This is a string that uniquely identifies this peptide in this reported peptide (e.g. "1" and "2" for cross-linked peptides.) IMPORTANT: If this file contains separate PSM-level annotations for each peptide in a single cross-linked PSM (e.g. a XCorr for each linked peptide), then this attribute is required, and is referenced when reporting the PSM-level score/annotation for each peptide. </xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="matched_proteins">
		<xs:annotation>
			<xs:documentation>Contains the protein sequences matched by peptides found in the experiment, and associated names, descriptions, and taxonomy IDs for those proteins. This will typically come from the FASTA file used to search the data. Only the proteins matched by the data should be included, not all the proteins in the FASTA file.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="protein" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>A protein found in the experiment.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="protein">
		<xs:annotation>
			<xs:documentation>A protein found in the experiment.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="protein_annotation" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>Annotations for this sequence from the sequence database used to search the data. Since there may be multiple annotations for a single sequence (as separated by control-A characters in a FASTA file), each annotation should appear hear associated with a single protein.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element name="protein_isotope_labels" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Optional. If this is an isotopically labelled protein sequence, the isotope label for the protein should be listed here. Note that non-labeled and labeld proteins should be separate protein entries eventhough their protein sequences are the same.</xs:documentation>
					</xs:annotation>
					<xs:complexType>
						<xs:sequence>
							<xs:element name="protein_isotope_label" minOccurs="0">
								<xs:annotation>
									<xs:documentation>The isotope label for this protein.</xs:documentation>
								</xs:annotation>
								<xs:complexType>
									<xs:attribute name="label" use="required">
										<xs:annotation>
											<xs:documentation>The isotope label. Should be one of: 13C, 15N, 18O, or 2H</xs:documentation>
										</xs:annotation>
									</xs:attribute>
								</xs:complexType>
							</xs:element>
						</xs:sequence>
					</xs:complexType>
				</xs:element>
			</xs:sequence>
			<xs:attribute name="sequence" type="sequence_type" use="required">
				<xs:annotation>
					<xs:documentation>The amino acid sequence for this protein.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="protein_annotation">
		<xs:annotation>
			<xs:documentation>Annotations for this sequence from the sequence database used to search the data. Since there may be multiple annotations for a single sequence (as separated by control-A characters in a FASTA file), each annotation should appear hear associated with a single protein.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of this protein, used to report protein names in the ProXL web application. Typically, this is the name for a FASTA entry in a FASTA file in the form of: >NAME DESCRIPTION</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>The description of the protein.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="ncbi-taxonomy-id" type="xs:integer">
				<xs:annotation>
					<xs:documentation>The NCBI taxonomy ID for this protein. E.g. 9606 for human, or 4932 for S. cerevisiae.  Used for taxonomy-level filtering in the ProXL web application. If not present, or set to 0, the species name will appear as "Species Unknown" in the web application.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="modifications">
		<xs:annotation>
			<xs:documentation>The modifications identified for a peptide.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="modification" minOccurs="0" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="modification">
		<xs:annotation>
			<xs:documentation>A mass modification of a residue. E.g., a PTM</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="mass" type="mass" use="required">
				<xs:annotation>
					<xs:documentation>The minoisotopic mass, in daltons of the modification. E.g. 79.966331 for phosphorylation</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="position" type="peptide_position" use="optional">
				<xs:annotation>
					<xs:documentation>The position in the peptide (starting at 1) of the modification.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="is_n_terminal" type="xs:boolean">
				<xs:annotation>
					<xs:documentation>Set this to true if this mod is at the n-terminus of the peptide. Omit if this is not an n-terminal mod.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="is_c_terminal" type="xs:boolean">
				<xs:annotation>
					<xs:documentation>Set this to true if this mod is at the c-terminus of the peptide. Omit if this is not a c-terminal mod.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="isMonolink" type="xs:boolean">
				<xs:annotation>
					<xs:documentation>If this modification is a monolink, this is set to "true"</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="linked_position">
		<xs:annotation>
			<xs:documentation>A linked position in a peptide.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="position" type="peptide_position" use="required"/>
		</xs:complexType>
	</xs:element>
	<xs:element name="linked_positions">
		<xs:annotation>
			<xs:documentation>The list of linked positions in a peptide.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="linked_position" minOccurs="0" maxOccurs="2">
					<xs:annotation>
						<xs:documentation>A linked position in a peptide. A looplinked peptide will have two linked positions. A crosslinked peptide will have one (on each linked peptide). The first residue is position 1.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_psm_annotation_type">
		<xs:annotation>
			<xs:documentation>A numeric type of psm annotation (such as a score like a q-value for XCorr) that is meant to be used to filter PSMs based on quality of matches. Try to limit indexed annotation types to a minimum.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name for this score, such as "q-value". Displayed on web site.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>A description for what this score is. Will be shown in ProXL as a tooltip for this score.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="filter_direction" type="filter_direction_type" use="required">
				<xs:annotation>
					<xs:documentation>The direction a filterable annotation type is sorted in.  If set to "below", attributes with lower values are considered more significant (such as in the case of p-values). If set to "above", attributes with higher values are considered more significant (such as in the case of XCorr).</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="default_filter" type="xs:boolean">
				<xs:annotation>
					<xs:documentation>If present, and if set to true, this score will be used as a default PSM-level filter when viewing data. Use sparingly.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="default_filter_value" type="xs:decimal">
				<xs:annotation>
					<xs:documentation>This is the default cutoff value for this score when filtering data. If default_filter is set to true, this value must be set.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_psm_per_peptide_annotation_type">
		<xs:annotation>
			<xs:documentation>A numeric type of psm annotation (such as a score like a q-value for XCorr) that is meant to be used to filter PSMs based on quality of matches. Try to limit indexed annotation types to a minimum.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name for this score, such as "q-value". Displayed on web site.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>A description for what this score is. Will be shown in ProXL as a tooltip for this score.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="filter_direction" type="filter_direction_type" use="required">
				<xs:annotation>
					<xs:documentation>The direction a filterable annotation type is sorted in.  If set to "below", attributes with lower values are considered more significant (such as in the case of p-values). If set to "above", attributes with higher values are considered more significant (such as in the case of XCorr).</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_peptide_annotation_type">
		<xs:annotation>
			<xs:documentation>A numeric type of peptide annotation (such as a score like a q-value for XCorr) that is meant to be used to filter peptides based on quality of matches. Try to limit indexed annotation types to a minimum.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name for this score, such as "q-value". Displayed on web site.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>A description for what this score is. Will be shown in ProXL as a tooltip for this score.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="filter_direction" type="filter_direction_type" use="required">
				<xs:annotation>
					<xs:documentation>The direction a filterable annotation type is sorted in.  If set to "below", attributes with lower values are considered more significant (such as in the case of p-values). If set to "above", attributes with higher values are considered more significant (such as in the case of XCorr).</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="default_filter" type="xs:boolean">
				<xs:annotation>
					<xs:documentation>If present, and if set to true, this score will be used as a default peptide-level filter when viewing data. Use sparingly.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="default_filter_value" type="xs:decimal">
				<xs:annotation>
					<xs:documentation>This is the default cutoff value for this score when filtering data. If default_filter is set to true, this value must be set.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_annotation_type">
		<xs:annotation>
			<xs:documentation>An annotation of any type (numeric or string) that is not meant to be used as a possible filter for filtering out PSMs based on quality of matches. Example could include "scan number" or "charge"</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name for this score, such as "q-value". Displayed on web site.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>A description for what this score is. Will be shown in ProXL as a tooltip for this score.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_per_peptide_annotation_type">
		<xs:annotation>
			<xs:documentation>An annotation of any type (numeric or string) that is not meant to be used as a possible filter for filtering out PSMs based on quality of matches. Example could include "scan number" or "charge"</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name for this score, such as "q-value". Displayed on web site.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>A description for what this score is. Will be shown in ProXL as a tooltip for this score.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_peptide_annotation_type">
		<xs:annotation>
			<xs:documentation>An annotation of any type (numeric or string) that is not meant to be used as a possible filter for filtering out peptides based on quality of matches.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name for this score, such as "q-value". Displayed on web site.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="description" type="xs:string">
				<xs:annotation>
					<xs:documentation>A description for what this score is. Will be shown in ProXL as a tooltip for this score.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:simpleType name="filter_direction_type">
		<xs:annotation>
			<xs:documentation>The direction a filterable annotation type is sorted in.  If set to "below", attributes with lower values are considered more significant (such as in the case of p-values). If set to "above", attributes with higher values are considered more significant (such as in the case of XCorr).</xs:documentation>
		</xs:annotation>
		<xs:restriction base="xs:string">
			<xs:enumeration value="above"/>
			<xs:enumeration value="below"/>
		</xs:restriction>
	</xs:simpleType>
	<xs:simpleType name="sequence_type">
		<xs:annotation>
			<xs:documentation>Reported sequence strings must be of this type, includes only amino acid sequence</xs:documentation>
		</xs:annotation>
		<xs:restriction base="xs:string">
		</xs:restriction>
	</xs:simpleType>
	<xs:simpleType name="amino_acid_residue">
		<xs:annotation>
			<xs:documentation>The single letter code for an amino acid. E.g., "K" for lysine.</xs:documentation>
		</xs:annotation>
		<xs:restriction base="xs:string">
		</xs:restriction>
	</xs:simpleType>
	<xs:simpleType name="link_type">
		<xs:annotation>
			<xs:documentation>A peptide link type must be this type, limits to 'crosslink' or 'looplink' or 'unlinked' 'unlinked means neither a crosslink nor looplink</xs:documentation>
		</xs:annotation>
		<xs:restriction base="xs:string">
			<xs:enumeration value="crosslink"/>
			<xs:enumeration value="looplink"/>
			<xs:enumeration value="unlinked"/>
		</xs:restriction>
	</xs:simpleType>
	<xs:simpleType name="protein_terminus_designation">
		<xs:annotation>
			<xs:documentation>The terminus of a protein, either "n" or "c".</xs:documentation>
		</xs:annotation>
		<xs:restriction base="xs:string">
			<xs:enumeration value="n"/>
			<xs:enumeration value="c"/>
		</xs:restriction>
	</xs:simpleType>
	<xs:simpleType name="peptide_position">
		<xs:annotation>
			<xs:documentation>A position in a peptide, where 1 is the first position.</xs:documentation>
		</xs:annotation>
		<xs:restriction base="xs:positiveInteger">
			<xs:minInclusive value="1"/>
		</xs:restriction>
	</xs:simpleType>
	<xs:simpleType name="mass">
		<xs:annotation>
			<xs:documentation>Mass represented in daltons</xs:documentation>
		</xs:annotation>
		<xs:restriction base="xs:decimal"/>
	</xs:simpleType>
	<xs:element name="filterable_psm_annotation_types">
		<xs:annotation>
			<xs:documentation>The collection of numeric annotation types (such as scores) that may be associated with a PSM that are meant to be used to filter based on quality</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="filterable_psm_annotation_type" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_psm_per_peptide_annotation_types">
		<xs:annotation>
			<xs:documentation>The collection of numeric annotation types (such as scores) that may be associated with a PSM that are meant to be used to filter based on quality</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="filterable_psm_per_peptide_annotation_type" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_peptide_annotation_types">
		<xs:annotation>
			<xs:documentation>The collection of numeric annotation types (such as scores) that may be associated with a reported peptide that are meant to be used to filter based on quality</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="filterable_peptide_annotation_type" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_annotation_types">
		<xs:annotation>
			<xs:documentation>The collection of annotation types (string or numeric) associated with PSMs that are descriptive or not meant to be used to filter based on quality</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="descriptive_psm_annotation_type" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_per_peptide_annotation_types">
		<xs:annotation>
			<xs:documentation>The collection of annotation types (string or numeric) associated with PSMs that are descriptive or not meant to be used to filter based on quality</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="descriptive_psm_per_peptide_annotation_type" minOccurs="0" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_peptide_annotation_types">
		<xs:annotation>
			<xs:documentation>The collection of annotation types (string or numeric) associated with PSMs that are descriptive or not meant to be used to filter based on quality</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="descriptive_peptide_annotation_type" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>An annotation of any type (numeric or string) that is not meant to be used as a possible filter for filtering out peptides based on quality of matches. Examples include calculated mass or validation status.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_psm_annotation">
		<xs:annotation>
			<xs:documentation>A numeric annotation (usually a score) assigned to a peptide spectrum match (PSM) that is meant to describe the quality of the match and may be used to filter out poor matches. Examples include p-value, q-value, XCorr</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the search program, must be found in the search_programs element</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the psm annotation type (e.g., "q-value"). Must be found in the possible score types for the above search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="value" type="xs:decimal" use="required">
				<xs:annotation>
					<xs:documentation>The value (ie, score) assigned for the given annotationtype from the given search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_psm_per_peptide_annotation">
		<xs:annotation>
			<xs:documentation>A numeric annotation (usually a score) assigned to a peptide spectrum match (PSM) that is meant to describe the quality of the match and may be used to filter out poor matches. Examples include p-value, q-value, XCorr</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the search program, must be found in the search_programs element</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the psm annotation type (e.g., "q-value"). Must be found in the possible score types for the above search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="value" type="xs:decimal" use="required">
				<xs:annotation>
					<xs:documentation>The value (ie, score) assigned for the given annotationtype from the given search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_annotation">
		<xs:annotation>
			<xs:documentation>A descriptive annotation (numeric or string) assigned to a peptide spectrum match (PSM) that is not meant to describe the quality of the match or filter out poor matches. Examples include scan number of precursor charge state.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the search program, must be found in the search_programs element</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the psm annotation type (e.g., "q-value"). Must be found in the possible score types for the above search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="value" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The value (ie, score) assigned for the given annotationtype from the given search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_per_peptide_annotation">
		<xs:annotation>
			<xs:documentation>A descriptive annotation (numeric or string) assigned to a peptide spectrum match (PSM) that is not meant to describe the quality of the match or filter out poor matches. Examples include scan number of precursor charge state.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the search program, must be found in the search_programs element</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of the psm annotation type (e.g., "q-value"). Must be found in the possible score types for the above search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="value" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The value (ie, score) assigned for the given annotationtype from the given search program</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_psm_annotations">
		<xs:annotation>
			<xs:documentation>The numeric annotations (usually scores of some type) assigned to the peptide spectrum match (PSM) that are meant to describe the quality of the match and may be used to filter out poor matches. Examples include p-value, q-value, XCorr</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="filterable_psm_annotation" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="filterable_psm_per_peptide_annotations">
		<xs:annotation>
			<xs:documentation>The numeric annotations (usually scores of some type) assigned to the peptide spectrum match (PSM) that are meant to describe the quality of the match and may be used to filter out poor matches. Examples include p-value, q-value, XCorr</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="filterable_psm_per_peptide_annotation" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_annotations">
		<xs:annotation>
			<xs:documentation>The descriptive annotations (numeric or string) assigned to the peptide spectrum match (PSM) that are not meant to describe the quality of the match or filter out poor matches. Examples include scan number of precursor charge state.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="descriptive_psm_annotation" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>A descriptive annotation (numeric or string) assigned to a peptide spectrum match (PSM) that is not meant to describe the quality of the match or filter out poor matches.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="descriptive_psm_per_peptide_annotations">
		<xs:annotation>
			<xs:documentation>The descriptive annotations (numeric or string) assigned to the peptide spectrum match (PSM) that are not meant to describe the quality of the match or filter out poor matches. Examples include scan number of precursor charge state.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="descriptive_psm_per_peptide_annotation" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>A descriptive annotation (numeric or string) assigned to a peptide spectrum match (PSM) that is not meant to describe the quality of the match or filter out poor matches.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="psm">
		<xs:annotation>
			<xs:documentation>A PSM</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="filterable_psm_annotations"/>
				<xs:element ref="descriptive_psm_annotations" minOccurs="0">
					<xs:annotation>
						<xs:documentation>The descriptive annotations (numeric or string) assigned to the peptide spectrum match (PSM) that are not meant to describe the quality of the match or filter out poor matches.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element name="per_peptide_annotations" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Optional. Must use when there are separate data to report for each peptide in a cross-link PSM. Examples include cleavable cross-linker experiments or whenever a separate score score is provided for each peptide in a cross-link (e.g. an XCorr for each separate peptide ID).</xs:documentation>
					</xs:annotation>
					<xs:complexType>
						<xs:sequence>
							<xs:element name="psm_peptide" maxOccurs="2">
								<xs:annotation>
									<xs:documentation>Each distinct peptide found in this PSM. For cross-linked PSMs, there will be 2 psm_peptide elements.</xs:documentation>
								</xs:annotation>
								<xs:complexType>
									<xs:sequence>
										<xs:element ref="filterable_psm_per_peptide_annotations">
											<xs:annotation>
												<xs:documentation>The numeric annotations (usually scores of some type) assigned to a specific peptide that is part of a peptide spectrum match (PSM) that are meant to describe the quality of the match and may be used to filter out poor matches. Examples include p-value, q-value, XCorr</xs:documentation>
											</xs:annotation>
										</xs:element>
										<xs:element ref="descriptive_psm_per_peptide_annotations" minOccurs="0">
											<xs:annotation>
												<xs:documentation>The descriptive annotations (numeric or string) assigned to a specific peptide that is part of the peptide spectrum match (PSM) that are not meant to describe the quality of the match or filter out poor matches. Examples include scan number of precursor charge state.</xs:documentation>
											</xs:annotation>
										</xs:element>
									</xs:sequence>
									<xs:attribute name="unique_id" use="required">
										<xs:annotation>
											<xs:documentation>The unique id given to this peptide for this reported_peptide (e.g. "1" or "2". Recall that a reported_peptide is the identified ion, so will contain two separate peptides in the case of a cross-linked peptide. See the peptide element under the peptides element for a reported_peptide.</xs:documentation>
										</xs:annotation>
									</xs:attribute>
									<xs:attribute name="scan_number" type="xs:positiveInteger">
										<xs:annotation>
											<xs:documentation>(Optional) The scan number used to look up the scan in the given file. For cleavable cross-linker data, this will likely be the scan number of the MS3 scan for this peptide.</xs:documentation>
										</xs:annotation>
									</xs:attribute>
									<xs:attribute name="scan_file_name" type="xs:string">
										<xs:annotation>
											<xs:documentation>(Optional) The filename of the file in which this scan may be found. </xs:documentation>
										</xs:annotation>
									</xs:attribute>
									<xs:attribute name="linker_mass" type="mass">
										<xs:annotation>
											<xs:documentation>(Optional) Must be included for a cleavable cross-linker experiment. This is the mass of the side of the cross-linker that is bound to this peptide.</xs:documentation>
										</xs:annotation>
									</xs:attribute>
								</xs:complexType>
							</xs:element>
						</xs:sequence>
					</xs:complexType>
				</xs:element>
			</xs:sequence>
			<xs:attribute name="scan_file_name" type="xs:string">
				<xs:annotation>
					<xs:documentation>(Optional) The filename of the file in which this scan may be found. Will be used to locate the spectrum associated with this PSM when data are uploaded.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="scan_number" type="xs:positiveInteger">
				<xs:annotation>
					<xs:documentation>(Optional) The scan number used to look up the scan in the given file. Will be used to locate the spectrum associated with this PSM when data are uploaded.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="precursor_charge" type="xs:positiveInteger" use="required">
				<xs:annotation>
					<xs:documentation>The predicted charge of the precursor ion. Used for accurately annotating spectra in web interface.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="precursor_retention_time" type="xs:decimal">
				<xs:annotation>
					<xs:documentation>(Optional) The retention time (in seconds) of the precursor ion. If not present, the retention times from the optionally uploaded mz data files (e.g. mzML will be used).</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="precursor_m_z" type="xs:decimal">
				<xs:annotation>
					<xs:documentation>(Optional) The observed m/z of the precursor ion. If not present, the measured m/z from the optionally uploaded mz data files (e.g. mzML will be used).</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="linker_mass" type="mass">
				<xs:annotation>
					<xs:documentation>(Optional) If this PSM identified a looplinked peptide or crosslinked peptidie pair, this is the mass used for the crosslinker in this identification. Used for accurately annotating spectra in web interface.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="psms">
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="psm" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="search_programs">
		<xs:annotation>
			<xs:documentation>The search program(s) used to generate the PSMs and associated scores in this search. More than one search program may be used, as in the case of using Percolator to assign q-values for data search with Kojak. Percolator and Kojak would each have an entry here. However, all PSMs must be annotated by all search programs and all search programs must refer to reported peptides using the same string.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_program" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="linker">
		<xs:annotation>
			<xs:documentation>A chemical crosslinker used in this experiment.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="monolink_masses" minOccurs="0">
					<xs:annotation>
						<xs:documentation>The list of possible masses that monolinks may be reported as having in this search. </xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="crosslink_masses" minOccurs="0">
					<xs:annotation>
						<xs:documentation>The list of possible masses of this linker in this experiment. Required if cross-links or loop-links were searched for and if spectra are being imported.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="linked_ends" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Define where each end of the cross-linker may bind with proteins.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
			<xs:attribute name="name" type="xs:string" use="required">
				<xs:annotation>
					<xs:documentation>The name of this crosslinker (e.g., "dss" or "edc").</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="spacer_arm_length" type="xs:decimal">
				<xs:annotation>
					<xs:documentation>Optional. The length, in angstroms, of the spacer arm of the cross-linker. E.g., 10.3 for DSSO</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="monolink_mass">
		<xs:complexType>
			<xs:attribute name="mass" type="mass" use="required">
				<xs:annotation>
					<xs:documentation>The minoisotopic mass, in daltons.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="crosslink_mass">
		<xs:complexType>
			<xs:attribute name="mass" type="mass" use="required">
				<xs:annotation>
					<xs:documentation>The minoisotopic mass, in daltons.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="chemical_formula" type="xs:string">
				<xs:annotation>
					<xs:documentation>Optional. The chemical formula corresponding to this mass for this cross-linker. E.g., C6H6O3S for DSSO. Required for some reports in proxl.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="monolink_masses">
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="monolink_mass" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>A mass used for this linker in this search when only one end of the linker is bound to a peptide. Use the same precision or better as when monolinks are reported as dynamic mods on peptides.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="crosslink_masses">
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="crosslink_mass" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>A full-length mass used for this linker in this search when both ends are bound to peptides--such as in crosslinks or looplinks</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="cleaved_crosslink_mass" minOccurs="0" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="linkers">
		<xs:annotation>
			<xs:documentation>The chemical linkers used in this experiment. There must be at least one.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="linker" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="configuration_file">
		<xs:annotation>
			<xs:documentation>A configuration or properties file associated with the search. E.g. Kojak.conf</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="file_content"/>
			</xs:sequence>
			<xs:attribute name="search_program" use="required"/>
			<xs:attribute name="file_name" use="required"/>
		</xs:complexType>
	</xs:element>
	<xs:element name="file_content" type="xs:base64Binary">
		<xs:annotation>
			<xs:documentation>The contents of a file, stored as base64</xs:documentation>
		</xs:annotation>
	</xs:element>
	<xs:element name="configuration_files">
		<xs:annotation>
			<xs:documentation>The configuration files associated with the search.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="configuration_file" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="static_modification">
		<xs:annotation>
			<xs:documentation>A mass shift that is always applied to the given amino acid when calculating theoretical peptide or ion masses</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="amino_acid" use="required">
				<xs:annotation>
					<xs:documentation>The one-letter code for the amino acid with the static modification.</xs:documentation>
				</xs:annotation>
				<xs:simpleType>
					<xs:restriction base="xs:string">
						<xs:length value="1"/>
					</xs:restriction>
				</xs:simpleType>
			</xs:attribute>
			<xs:attribute name="mass_change" type="mass" use="required">
				<xs:annotation>
					<xs:documentation>The mass shift applied to all instances of the above amino acid.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="static_modifications">
		<xs:annotation>
			<xs:documentation>A collection of mass shifts that are always applied to the respective amino acids when calculating theoretical peptide or ion masses</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="static_modification" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="decoy_labels">
		<xs:annotation>
			<xs:documentation>A list of decoy labels (strings that are prefixes on protein names that indicate the given protein is a decoy (e.g., "DECOY_")</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="decoy_label" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="decoy_label">
		<xs:annotation>
			<xs:documentation>A decoy label is used to identify a protein to which a peptide is matched during protein inference as a decoy protein (used to calculate FDR statistics). Matches to decoy proteins will be ignored and not inserted into the database as a matched protein for peptides. Example: <decoy_label prefix="DECOY_"/>
			</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="prefix" type="xs:string">
				<xs:annotation>
					<xs:documentation>Any protein name beginning with the supplied string will be considered a decoy protein and ignored.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="search_program_info">
		<xs:annotation>
			<xs:documentation>Information relating to the search program(s) used to generate the peptide spectrum matches and scores/annotations for PSMs and peptides.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_programs"/>
				<xs:element ref="annotation_cutoffs_on_import" minOccurs="0">
					<xs:annotation>
						<xs:documentation>DEPRECATED - Will be ignored by proxl importer. May be removed in future schema release.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="default_visible_annotations"/>
				<xs:element ref="annotation_sort_order" minOccurs="0">
					<xs:annotation>
						<xs:documentation>The sort order applied to filterable annotations when presenting the default listing of PSMs or peptides. The order will be in the order the annotations appear here. E.g. if "q-value" is listed first and "p-value" is listed next, the items will be sorted on q-value first, then, sorted on p-value in the case of ties. Note: all annotation types listed here must be filterable. If this element is not present, no default sorting will be applied to listed PSMs or peptides in ProXL.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="annotation_cutoffs_on_import">
		<xs:annotation>
			<xs:documentation>The cutoffs used to exclude data while importing.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="psm_annotation_cutoffs_on_import">
					<xs:annotation>
						<xs:documentation>DEPRECATED - Will be ignored by proxl importer. May be removed in future schema release.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="reported_peptide_annotation_cutoffs_on_import" minOccurs="0">
					<xs:annotation>
						<xs:documentation>DEPRECATED - Will be ignored by proxl importer. May be removed in future schema release.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="psm_annotation_cutoffs_on_import">
		<xs:annotation>
			<xs:documentation>The PSM-level annotation cutoffs used to exclude data while importing.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_annotation_cutoff" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="reported_peptide_annotation_cutoffs_on_import">
		<xs:annotation>
			<xs:documentation>The peptide-level annotation cutoffs used to exclude data while importing.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_annotation_cutoff" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="search_annotation_cutoff">
		<xs:annotation>
			<xs:documentation>A reference to an annotation in a search program.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" use="required">
				<xs:annotation>
					<xs:documentation>The name of a search program, must match a name for a search program in this document. E.g., "kojak"</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" use="required">
				<xs:annotation>
					<xs:documentation>The name of the annotation, must match a name of an annotation for the search_program. E.g., "q-value"</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="cutoff_value" type="xs:decimal" use="required">
				<xs:annotation>
					<xs:documentation>This is the cutoff value for this score for excluding data while importing into Proxl.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="default_visible_annotations">
		<xs:annotation>
			<xs:documentation>The annotations from the search program(s) that are visible by default in the ProXL web app.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="visible_psm_annotations"/>
				<xs:element ref="visible_psm_per_peptide_annotations" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Optional. The PSM-level annotations that are associated individual with each peptide (for example, each separate peptide in a cross-link) that should be visible by default when viewing PSM-level data.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="visible_reported_peptide_annotations" minOccurs="0">
					<xs:annotation>
						<xs:documentation>Optional. The peptide-level annotations that are visible by default.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="visible_psm_annotations">
		<xs:annotation>
			<xs:documentation>The PSM-level annotations that are visible by default when viewing PSM-level data.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_annotation" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="visible_psm_per_peptide_annotations">
		<xs:annotation>
			<xs:documentation>The PSM-level annotations that are visible by default when viewing PSM-level data.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_annotation" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="visible_reported_peptide_annotations">
		<xs:annotation>
			<xs:documentation>The peptide-level annotations that are visible by default.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_annotation" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="search_annotation">
		<xs:annotation>
			<xs:documentation>A reference to an annotation in a search program.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="search_program" use="required">
				<xs:annotation>
					<xs:documentation>The name of a search program, must match a name for a search program in this document. E.g., "kojak"</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="annotation_name" use="required">
				<xs:annotation>
					<xs:documentation>The name of the annotation, must match a name of an annotation for the search_program. E.g., "q-value"</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="annotation_sort_order">
		<xs:annotation>
			<xs:documentation>The sort order applied to filterable annotation when presenting the default listing of PSMs or peptides. The order will be in the order the annotations appear here. E.g. if "q-value" is listed first and "p-value" is listed next, the items will be sorted on q-value first, then, sorted on p-value in the case of ties. Note: all annotation types listed here must be filterable. If this element is not present, no default sorting will be applied to listed PSMs or peptides in ProXL.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="psm_annotation_sort_order" minOccurs="0"/>
				<xs:element ref="reported_peptide_annotation_sort_order" minOccurs="0"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="psm_annotation_sort_order">
		<xs:annotation>
			<xs:documentation>The sort order applied to filterable PSM annotations when presenting the default listing of PSMs. The order will be in the order the annotations appear here. E.g. if "q-value" is listed first and "p-value" is listed next, the PSMs will be sorted on q-value first, then, sorted on p-value in the case of ties. Note: all annotation types listed here must be filterable. If this element is not present, no default sorting will be applied to listed PSMs.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_annotation" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="reported_peptide_annotation_sort_order">
		<xs:annotation>
			<xs:documentation>The sort order applied to filterable peptide annotations when presenting the default listing of peptides. The order will be in the order the annotations appear here. E.g. if "q-value" is listed first and "p-value" is listed next, the peptides will be sorted on q-value first, then, sorted on p-value in the case of ties. Note: all annotation types listed here must be filterable. If this element is not present, no default sorting will be applied to listed peptides.</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="search_annotation" maxOccurs="unbounded"/>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="cleaved_crosslink_mass">
		<xs:annotation>
			<xs:documentation>If the cross-linker cleavable, a mass to be used as a cleaved product of this cross-linker. Ie, the mass of each side of the cross-linker bound to a peptide after cleavage. A cleavable linkers will usually have two of these (but not always).</xs:documentation>
		</xs:annotation>
		<xs:complexType>
			<xs:attribute name="mass" type="mass" use="required">
				<xs:annotation>
					<xs:documentation>The minoisotopic mass, in daltons.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="chemical_formula" type="xs:string">
				<xs:annotation>
					<xs:documentation>Optional. The chemical formula corresponding to this mass for this cross-linker. E.g., C3H2O for one of the cleaved ends of DSSO.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
	<xs:element name="linked_ends">
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="linked_end" minOccurs="2" maxOccurs="2">
					<xs:annotation>
						<xs:documentation>Define where one end of the cross-linker may react with a protein. There must be two of these elements, even if both sides are the same. This must contain at least one residues element or protein_termini element.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="linked_end">
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="residues" minOccurs="0">
					<xs:annotation>
						<xs:documentation>A list of amino acid residues with which this end of the cross-linker may bind.</xs:documentation>
					</xs:annotation>
				</xs:element>
				<xs:element ref="protein_termini" minOccurs="0">
					<xs:annotation>
						<xs:documentation>A list of protein terminal positions to which this cross-linker can bind. E.g. 0 or 1 residues away from n-terminus.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="residue" type="amino_acid_residue"/>
	<xs:element name="residues">
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="residue" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>The single letter code for the residue with which this end of the crosslinker may bind. E.g., K</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="protein_termini">
		<xs:complexType>
			<xs:sequence>
				<xs:element ref="protein_terminus" maxOccurs="unbounded">
					<xs:annotation>
						<xs:documentation>A protein terminus, either n or c, and the number of residues away from that terminus the linker can bind.</xs:documentation>
					</xs:annotation>
				</xs:element>
			</xs:sequence>
		</xs:complexType>
	</xs:element>
	<xs:element name="protein_terminus">
		<xs:complexType>
			<xs:attribute name="terminus_end" type="protein_terminus_designation" use="required">
				<xs:annotation>
					<xs:documentation>The terminal end, either "n" or "c".</xs:documentation>
				</xs:annotation>
			</xs:attribute>
			<xs:attribute name="distance_from_terminus" type="xs:nonNegativeInteger" use="required">
				<xs:annotation>
					<xs:documentation>The distance from the given terminus. 0 indicates the terminus itself. 1 indicates one position removed from the terminus. E.g., in the protein PEPTIDE, a distance of 0 from the c terminus would be E, a distance of 1 would be D.</xs:documentation>
				</xs:annotation>
			</xs:attribute>
		</xs:complexType>
	</xs:element>
</xs:schema>
"""



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import io
import zipfile
import pandas as pd
from tqdm import tqdm

from ..data import check_input
from ..data import create_csm
from ..data import create_parser_result
from ..constants import AMINO_ACIDS
from ..constants import MODIFICATIONS
from ..constants import MEROX_MODIFICATION_MAPPING
from .util import __serialize_pandas_series

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List


MEROX_COLNAMES = [
    "Score",
    "m/z",
    "Charge",
    "M+H+",
    "Calculated Mass",
    "Deviation in ppm",
    "Peptide 1",
    "Protein 1",
    "From",
    "To",
    "Peptide2",
    "Protein 2",
    "From.1",
    "To.1",
    "Scan number",
    "is Selected in Table",
    "Candidate identifier",
    "Folder Number",
    "Retention time in sec",
    "miscellaneous",
    "best linkage position peptide 1",
    "best linkage position peptide 2",
    "All linkage positions",
    "Spectrum UUID",
    "local False discovery rate",
    "shortest distance in pdb",
    "Light/Heavy(1/2)",
    "pepScore1",
    "pepScore2",
    "xLinkScore",
    "resultId",
    "MS1intensity",
    "finalScoreComponent",
]


def __read_merox_file(
    file: str | BinaryIO, sep: str = ";", decimal: str = "."
) -> pd.DataFrame:
    r"""Helper function to read MeroX files into pandas DataFrames.

    Reads MeroX files into pandas DataFrames independent of input format. Accepts
    both ``.csv`` and ``.zhrm`` files.

    Parameters
    ----------
    file : str, or file stream
        The name/path of the MeroX result file or a file-like object/stream.
    sep : str, default = ";"
        Seperator used in the ``.csv`` or ``.zhrm`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.

    Returns
    -------
    pd.DataFrame
        The MeroX result file as a pandas DataFrame.

    Notes
    -----
    This function should not be called directly, it is called from ``read_merox()``.
    """
    # safety
    if not isinstance(file, str):
        file.seek(0)
    # this shifts file pointer
    if zipfile.is_zipfile(file):
        with zipfile.ZipFile(file, "r") as f:
            return pd.read_csv(
                io.BytesIO(f.read("Result.csv")),
                header=None,
                names=MEROX_COLNAMES,
                sep=sep,
                decimal=decimal,
                low_memory=False,
            )
    if not isinstance(file, str):
        file.seek(0)
    return pd.read_csv(file, sep=sep, decimal=decimal, low_memory=False)


def __get_merox_sequence(
    sequence: str,
    parse_modifications: bool = True,
    modifications: Dict[str, Dict[str, Any]] = MEROX_MODIFICATION_MAPPING,
) -> str:
    r"""Helper funtion to read the MeroX peptide sequence.

    Reads the MeroX peptide sequence and removes prefix and suffix square brackets,
    then replaces modification symbols with their respective amino acids. If
    ``parse_modifications = True`` non-resolved symbols will raise an error,
    otherwise the symbol will be used in the sequence - even if it is not a valid
    amino acid sequence anymore.

    Parameters
    ----------
    sequence : str
        The MeroX sequence e.g. from column "Peptide 1" or "Peptide2".
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications: dict of str, dict of str, any, default = ``constants.MEROX_MODIFICATION_MAPPING``
        Mapping of modification symbols to their amino acids and modifications. Please refer to
        ``constants.MEROX_MODIFICATION_MAPPING`` for examples.

    Returns
    -------
    str
        The parsed sequence.

    Raises
    ------
    KeyError
        If one of the symbols in the sequence could not be resolved/an unknown modification is encountered.

    Notes
    -----
    This function should not be called directly, it is called from ``read_merox()``.
    """
    parsed_seq = ""
    for amino_acid in sequence.lstrip("[").rstrip("]").strip():
        if amino_acid in modifications:
            parsed_seq += modifications[amino_acid]["Amino Acid"]
        elif amino_acid in AMINO_ACIDS:
            parsed_seq += amino_acid
        else:
            if parse_modifications:
                raise KeyError(
                    f"Key {amino_acid} not found in parameter 'modifications'. Are you missing a modification?"
                )
            else:
                parsed_seq += amino_acid
    return parsed_seq


def __get_merox_modifications(
    sequence: str,
    crosslink_position: int,
    crosslinker: str,
    crosslinker_mass: float,
    modifications: Dict[str, Dict[str, Any]] = MEROX_MODIFICATION_MAPPING,
) -> Dict[int, Tuple[str, float]]:
    r"""Helper function to read modifications from a MeroX sequence.

    Parses post-translational-modifications from the given MeroX sequence.

    Parameters
    ----------
    sequence : str
        The MeroX sequence e.g. from column "Peptide 1" or "Peptide2".
    crosslink_position : int
        Position of the crosslinker in the sequence (1-based).
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.
    modifications: dict of str, dict of str, any, default = ``constants.MEROX_MODIFICATION_MAPPING``
        Mapping of modification symbols to their amino acids and modifications. Please refer to
        ``constants.MEROX_MODIFICATION_MAPPING`` for examples.

    Returns
    -------
    dict of int, tuple of str, float
        The pyXLMS specfic modification representation of the parsed modifications.

    Raises
    ------
    RuntimeError
        If multiple modifications on the same residue are parsed.
    KeyError
        If an unknown modification is encountered.

    Notes
    -----
    This function should not be called directly, it is called from ``read_merox()``.
    """
    parsed_modifications = {crosslink_position: (crosslinker, crosslinker_mass)}
    for i, amino_acid in enumerate(sequence.lstrip("[").rstrip("]").strip()):
        if amino_acid in modifications:
            if i + 1 not in parsed_modifications:
                parsed_modifications[i + 1] = modifications[amino_acid]["Modification"]
            else:
                raise RuntimeError(f"Modification at position {i + 1} already exists!")
        elif amino_acid in AMINO_ACIDS:
            pass
        else:
            raise KeyError(
                f"Key {amino_acid} not found in parameter 'modifications'. Are you missing a modification?"
            )
    return parsed_modifications


def __get_merox_position(position_str: str) -> int:
    r"""Helper function to extract the peptide crosslink position from MeroX.

    Parameters
    ----------
    position_str : str
        The position string from MeroX e.g. from column "best linkage position peptide 1".

    Returns
    -------
    int
        The parsed peptide crosslink position.

    Raises
    ------
    RuntimeError
        If the position could not be parsed.

    Notes
    -----
    This function should not be called directly, it is called from ``read_merox()``.
    """
    position = None
    try:
        position = int(position_str[1:])
    except Exception as _e:
        pass
    if position is None:
        raise RuntimeError(f"Could not parse position from {position_str}!")
    return position


def __get_merox_protein(proteins: str) -> List[str]:
    r"""Helper function to extract the protein accession from MeroX.

    Parses the leading protein accession from the MeroX protein string. Additional
    proteins are not parsed as they do not have corresponding protein crosslink
    positions.

    Parameters
    ----------
    proteins : str
        The protein string from MeroX e.g. from column "Protein 1".

    Returns
    -------
    list of str
        A list containing the single parsed protein accession.

    Notes
    -----
    This function should not be called directly, it is called from ``read_merox()``.
    """
    return [proteins.split("(>")[0].strip().lstrip(">")]


def __get_merox_scan_number(scan_nr_and_file: str) -> int:
    r"""Helper function to parse the scan number from MeroX.

    Parses the scan number from the MeroX scan number string. Assumes that scan
    number and spectrum file are delimted by the tilde (wave) sign.

    Parameters
    ----------
    scan_nr_and_file : str
        The scan number string from MeroX e.g. from the column "Scan number".

    Returns
    -------
    int
        The parsed scan number.

    Raises
    ------
    RuntimeError
        If the scan number could not be parsed.

    Notes
    -----
    This function should not be called directly, it is called from ``read_merox()``.
    """
    scan_nr = None
    try:
        scan_nr = int(scan_nr_and_file.split("~")[0])
    except Exception as _e:
        pass
    if scan_nr is None:
        raise RuntimeError(f"Could not scan number from {scan_nr_and_file}!")
    return scan_nr


def __get_merox_spectrum_file(scan_nr_and_file: str) -> str:
    r"""Helper function to parse the spectrum file from MeroX.

    Parses the spectrum file from the MeroX scan number string. Assumes that scan
    number and spectrum file are delimted by the tilde (wave) sign.

    Parameters
    ----------
    scan_nr_and_file : str
        The scan number string from MeroX e.g. from the column "Scan number".

    Returns
    -------
    str
        The parsed spectrum file.

    Raises
    ------
    RuntimeError
        If the spectrum file could not be parsed.

    Notes
    -----
    This function should not be called directly, it is called from ``read_merox()``.
    """
    spectrum_file = None
    try:
        spectrum_file = scan_nr_and_file.split("~")[1]
    except Exception as _e:
        pass
    if spectrum_file is None:
        raise RuntimeError(f"Could not spectrum file name from {scan_nr_and_file}!")
    return spectrum_file


def read_merox(
    files: str | List[str] | BinaryIO,
    crosslinker: str,
    crosslinker_mass: Optional[float] = None,
    decoy_prefix: str = "REV__",
    parse_modifications: bool = True,
    modifications: Dict[str, Dict[str, Any]] = MEROX_MODIFICATION_MAPPING,
    sep: str = ";",
    decimal: str = ".",
) -> Dict[str, Any]:
    r"""Read a MeroX result file.

    Reads a MeroX crosslink-spectrum-matches result file in ``.csv`` or ``.zhrm`` format
    and returns a ``parser_result``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the MeroX result file(s) or a file-like object/stream.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float, or None, default = None
        Monoisotopic delta mass of the crosslink modification. If the crosslinker is
        defined in ``constants.MODIFICATIONS`` this can be omitted.
    decoy_prefix : str, default = "REV\_\_"
        The prefix that indicates that a protein is from the decoy database.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications: dict of str, dict of str, any, default = ``constants.MEROX_MODIFICATION_MAPPING``
        Mapping of modification symbols to their amino acids and modifications. Please refer to
        ``constants.MEROX_MODIFICATION_MAPPING`` for examples.
    sep : str, default = ";"
        Seperator used in the ``.csv`` or ``.zhrm`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslink-spectrum-matches.
    KeyError
        If the specified crosslinker could not be found/mapped.

    Notes
    -----
    Uses ``pepScore1`` as the score for the alpha peptide, ``pepScore2`` as the score of the
    beta peptide, and ``Score`` as the score of the crosslink-spectrum-match.

    Warnings
    --------
    MeroX only reports a single protein crosslink position per peptide, for ambiguous peptides
    only the crosslink position of the first matching protein is reported. All matching proteins can be
    retrieved via ``additional_information``, however not their corresponding crosslink positions. For this
    reason it is recommended to use ``transform.reannotate_positions()`` to correctly annotate all crosslink
    positions for all peptides if that is important for downstream analysis. Additionally, please note that
    target and decoy information is derived based off the protein accession and parameter ``decoy_prefix``.
    By default, MeroX only reports target matches that are above the desired FDR.

    Examples
    --------
    >>> from pyXLMS.parser import read_merox
    >>> csms_from_csv = read_merox(
    ...     "data/merox/XLpeplib_Beveridge_QEx-HFX_DSS_R1.csv", crosslinker="DSS"
    ... )

    >>> from pyXLMS.parser import read_merox
    >>> csms_from_zhrm = read_merox(
    ...     "data/merox/XLpeplib_Beveridge_QEx-HFX_DSS_R1.zhrm", crosslinker="DSS"
    ... )
    """
    ## check input
    _ok = check_input(crosslinker, "crosslinker", str)
    _ok = (
        check_input(crosslinker_mass, "crosslinker_mass", float)
        if crosslinker_mass is not None
        else True
    )
    _ok = check_input(decoy_prefix, "decoy_prefix", str)
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = check_input(modifications, "modifications", dict, dict)
    _ok = check_input(sep, "sep", str)
    _ok = check_input(decimal, "decimal", str)
    if crosslinker_mass is None:
        if crosslinker not in MODIFICATIONS:
            if parse_modifications:
                raise KeyError(
                    "Cannot infer crosslinker mass because crosslinker is unknown. "
                    "Please specify crosslinker mass manually!"
                )
            else:
                crosslinker_mass = 0.0
        else:
            crosslinker_mass = MODIFICATIONS[crosslinker]

    ## data structures
    csms = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    for input in inputs:
        ## reading data
        data = __read_merox_file(input, sep=sep, decimal=decimal)
        for i, row in tqdm(
            data.iterrows(),
            total=data.shape[0],
            desc="Reading MeroX CSMs...",
        ):
            # create crosslink
            csm = create_csm(
                peptide_a=__get_merox_sequence(
                    str(row["Peptide 1"]), parse_modifications, modifications
                ),
                modifications_a=__get_merox_modifications(
                    str(row["Peptide 1"]),
                    __get_merox_position(str(row["best linkage position peptide 1"])),
                    crosslinker,
                    crosslinker_mass,
                    modifications,
                )
                if parse_modifications
                else None,
                xl_position_peptide_a=__get_merox_position(
                    str(row["best linkage position peptide 1"])
                ),
                proteins_a=__get_merox_protein(str(row["Protein 1"])),
                xl_position_proteins_a=[
                    int(row["From"])
                    + __get_merox_position(str(row["best linkage position peptide 1"]))
                    - 1
                ],
                pep_position_proteins_a=[int(row["From"])],
                score_a=float(row["pepScore1"]),
                decoy_a=__get_merox_protein(str(row["Protein 1"]))[0].startswith(
                    decoy_prefix
                ),
                peptide_b=__get_merox_sequence(
                    str(row["Peptide2"]), parse_modifications, modifications
                ),
                modifications_b=__get_merox_modifications(
                    str(row["Peptide2"]),
                    __get_merox_position(str(row["best linkage position peptide 2"])),
                    crosslinker,
                    crosslinker_mass,
                    modifications,
                )
                if parse_modifications
                else None,
                xl_position_peptide_b=__get_merox_position(
                    str(row["best linkage position peptide 2"])
                ),
                proteins_b=__get_merox_protein(str(row["Protein 2"])),
                xl_position_proteins_b=[
                    int(row["From.1"])
                    + __get_merox_position(str(row["best linkage position peptide 2"]))
                    - 1
                ],
                pep_position_proteins_b=[int(row["From.1"])],
                score_b=float(row["pepScore2"]),
                decoy_b=__get_merox_protein(str(row["Protein 2"]))[0].startswith(
                    decoy_prefix
                ),
                score=float(row["Score"]),
                spectrum_file=__get_merox_spectrum_file(str(row["Scan number"])),
                scan_nr=__get_merox_scan_number(str(row["Scan number"])),
                charge=int(row["Charge"]),
                rt=float(row["Retention time in sec"]),
                im_cv=None,
                additional_information={
                    "source": __serialize_pandas_series(row),
                    "xLinkScore": row["xLinkScore"],
                    "Protein 1": row["Protein 1"],
                    "Protein 2": row["Protein 2"],
                    "MS1intensity": row["MS1intensity"],
                },
            )
            csms.append(csm)
    ## check results
    if len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="MeroX",
        csms=csms,
        crosslinks=None,
    )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS import data

    csm = data.create_csm_min(
        "PEPTIDE",
        1,
        "EDITPEP",
        3,
        "RUN_1",
        1,
    )
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    # alpha
    assert csm["alpha_peptide"] == "EDITPEP"
    assert csm["alpha_modifications"] is None
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert csm["alpha_proteins"] is None
    assert csm["alpha_proteins_crosslink_positions"] is None
    assert csm["alpha_proteins_peptide_positions"] is None
    assert csm["alpha_score"] is None
    assert csm["alpha_decoy"] is None
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert csm["beta_modifications"] is None
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] is None
    assert csm["beta_proteins_crosslink_positions"] is None
    assert csm["beta_proteins_peptide_positions"] is None
    assert csm["beta_score"] is None
    assert csm["beta_decoy"] is None
    # csm
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] is None
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] is None
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test2():
    from pyXLMS import data

    csm = data.create_csm_min("PEPTIDE", 1, "EDITPEP", 3, "RUN_1", 1, score=170.3)
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    # alpha
    assert csm["alpha_peptide"] == "EDITPEP"
    assert csm["alpha_modifications"] is None
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert csm["alpha_proteins"] is None
    assert csm["alpha_proteins_crosslink_positions"] is None
    assert csm["alpha_proteins_peptide_positions"] is None
    assert csm["alpha_score"] is None
    assert csm["alpha_decoy"] is None
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert csm["beta_modifications"] is None
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] is None
    assert csm["beta_proteins_crosslink_positions"] is None
    assert csm["beta_proteins_peptide_positions"] is None
    assert csm["beta_score"] is None
    assert csm["beta_decoy"] is None
    # csm
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] == pytest.approx(170.3)
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] is None
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test3():
    from pyXLMS import data

    csm = data.create_csm_min(
        "PEPTIDE",
        1,
        "EDITPEP",
        3,
        "RUN_1",
        1,
        modifications_a={1: (" Oxidation ", 15.994915)},
        proteins_a=["PROTEIN"],
        xl_position_proteins_a=[1],
        pep_position_proteins_a=[1],
        score_a=50.3,
        decoy_a=False,
        modifications_b={2: ("Oxidation", 15.994915)},
        proteins_b=["NIETORP", "PROTEIN"],
        xl_position_proteins_b=[7, 4],
        pep_position_proteins_b=[5, 2],
        score_b=170.3,
        decoy_b=False,
        score=170.3,
        charge=3,
        rt=23.4,
        im_cv=-50.0,
        additional_information={"m/z": 1337.1},
    )
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "full"
    # alpha
    assert csm["alpha_peptide"] == "EDITPEP"
    assert len(csm["alpha_modifications"]) == 1
    assert 2 in csm["alpha_modifications"]
    assert csm["alpha_modifications"][2][0] == "Oxidation"
    assert csm["alpha_modifications"][2][1] == pytest.approx(15.994915)
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert len(csm["alpha_proteins"]) == 2
    assert csm["alpha_proteins"][0] == "NIETORP"
    assert len(csm["alpha_proteins_crosslink_positions"]) == 2
    assert csm["alpha_proteins_crosslink_positions"][0] == 7
    assert len(csm["alpha_proteins_peptide_positions"]) == 2
    assert csm["alpha_proteins_peptide_positions"][0] == 5
    assert csm["alpha_score"] == pytest.approx(170.3)
    assert not csm["alpha_decoy"]
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert len(csm["beta_modifications"]) == 1
    assert 1 in csm["beta_modifications"]
    assert csm["beta_modifications"][1][0] == "Oxidation"
    assert csm["beta_modifications"][1][1] == pytest.approx(15.994915)
    assert csm["beta_peptide_crosslink_position"] == 1
    assert len(csm["beta_proteins"]) == 1
    assert csm["beta_proteins"][0] == "PROTEIN"
    assert len(csm["beta_proteins_crosslink_positions"]) == 1
    assert csm["beta_proteins_crosslink_positions"][0] == 1
    assert len(csm["beta_proteins_peptide_positions"]) == 1
    assert csm["beta_proteins_peptide_positions"][0] == 1
    assert csm["beta_score"] == pytest.approx(50.3)
    assert not csm["beta_decoy"]
    # csm
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(170.3)
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] == 3
    assert csm["retention_time"] == pytest.approx(23.4)
    assert csm["ion_mobility"] == pytest.approx(-50.0)
    assert csm["additional_information"]["m/z"] == pytest.approx(1337.1)



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS import data

    crosslink = data.create_crosslink_min(
        "PEPTIDE",
        1,
        "EDITPEP",
        3,
    )
    assert crosslink["data_type"] == "crosslink"
    assert crosslink["completeness"] == "partial"
    assert crosslink["alpha_peptide"] == "EDITPEP"
    assert crosslink["alpha_peptide_crosslink_position"] == 3
    assert crosslink["alpha_proteins"] is None
    assert crosslink["alpha_proteins_crosslink_positions"] is None
    assert crosslink["alpha_decoy"] is None
    assert crosslink["beta_peptide"] == "PEPTIDE"
    assert crosslink["beta_peptide_crosslink_position"] == 1
    assert crosslink["beta_proteins"] is None
    assert crosslink["beta_proteins_crosslink_positions"] is None
    assert crosslink["beta_decoy"] is None
    assert crosslink["crosslink_type"] == "inter"
    assert crosslink["score"] is None


def test2():
    from pyXLMS import data

    crosslink = data.create_crosslink_min("PEPTIDE", 1, "EDITPEP", 3, score=170.3)
    assert crosslink["data_type"] == "crosslink"
    assert crosslink["completeness"] == "partial"
    assert crosslink["alpha_peptide"] == "EDITPEP"
    assert crosslink["alpha_peptide_crosslink_position"] == 3
    assert crosslink["alpha_proteins"] is None
    assert crosslink["alpha_proteins_crosslink_positions"] is None
    assert crosslink["alpha_decoy"] is None
    assert crosslink["beta_peptide"] == "PEPTIDE"
    assert crosslink["beta_peptide_crosslink_position"] == 1
    assert crosslink["beta_proteins"] is None
    assert crosslink["beta_proteins_crosslink_positions"] is None
    assert crosslink["beta_decoy"] is None
    assert crosslink["crosslink_type"] == "inter"
    assert crosslink["score"] == pytest.approx(170.3)


def test3():
    from pyXLMS import data

    crosslink = data.create_crosslink_min(
        "PEPTIDE",
        1,
        "EDITPEP",
        3,
        proteins_a=["Protein"],
        xl_position_proteins_a=[15],
        decoy_a=True,
        proteins_b=["Protein"],
        xl_position_proteins_b=[37],
        decoy_b=False,
        score=170.3,
        additional_information={"charge": 3},
    )
    assert crosslink["data_type"] == "crosslink"
    assert crosslink["completeness"] == "full"
    assert crosslink["alpha_peptide"] == "EDITPEP"
    assert crosslink["alpha_peptide_crosslink_position"] == 3
    assert crosslink["alpha_proteins"] == ["Protein"]
    assert crosslink["alpha_proteins_crosslink_positions"] == [37]
    assert not crosslink["alpha_decoy"]
    assert crosslink["beta_peptide"] == "PEPTIDE"
    assert crosslink["beta_peptide_crosslink_position"] == 1
    assert crosslink["beta_proteins"] == ["Protein"]
    assert crosslink["beta_proteins_crosslink_positions"] == [15]
    assert crosslink["beta_decoy"]
    assert crosslink["crosslink_type"] == "intra"
    assert crosslink["score"] == pytest.approx(170.3)
    assert crosslink["additional_information"]["charge"] == 3



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest
import pandas as pd


CROSSLINKS_ALL = "data/_test/exporter/pyxlinkviewer/unique_links_all.xlsx"
CROSSLINKS_NSP8 = "data/_test/exporter/pyxlinkviewer/unique_links_nsp8.xlsx"
YHU_PDB = "data/_test/exporter/pyxlinkviewer/6YHU/6yhu.pdb"
YHU_RESULT_ALL = (
    "data/_test/exporter/pyxlinkviewer/6YHU/all/unique_links_all_crosslinks.txt"
)
YHU_RESULT_NSP8 = (
    "data/_test/exporter/pyxlinkviewer/6YHU/nsp8/unique_links_nsp8_crosslinks.txt"
)
JLT_PDB = "data/_test/exporter/pyxlinkviewer/7JLT/7jlt.pdb"
JLT_RESULT_NSP8 = (
    "data/_test/exporter/pyxlinkviewer/7JLT/nsp8/unique_links_nsp8_7jlt_crosslinks.txt"
)


def helper_get_crosslink_position(pep_seq: str) -> int:
    for i, aa in enumerate(pep_seq.strip()):
        if aa == "[":
            return i


def helper_get_crosslinks(f: str) -> pd.DataFrame:
    from pyXLMS.data import create_crosslink_min
    from pyXLMS.parser.util import format_sequence

    df = pd.read_excel(f)
    xls = list()
    for i, row in df.iterrows():
        xls.append(
            create_crosslink_min(
                format_sequence(row["Sequence A"]),
                helper_get_crosslink_position(row["Sequence A"]) + 1,
                format_sequence(row["Sequence B"]),
                helper_get_crosslink_position(row["Sequence B"]) + 1,
            )
        )
    return xls


def helper_get_dataframe_from_pyxlinkviewer(filename: str) -> pd.DataFrame:
    pyxlinkviewer = ""
    with open(filename, "r", encoding="utf-8") as f:
        pyxlinkviewer = f.read()
        f.close()
    pos_a = list()
    chain_a = list()
    pos_b = list()
    chain_b = list()
    for line in pyxlinkviewer.split("\n"):
        if len(line.strip()) > 0:
            xl = line.split("|")
            pos_a.append(int(xl[0]))
            chain_a.append(xl[1].strip())
            pos_b.append(int(xl[2]))
            chain_b.append(xl[3].strip())
    return pd.DataFrame(
        {"residue 1": pos_a, "chain 1": chain_a, "residue 2": pos_b, "chain 2": chain_b}
    )


def test1():
    import os
    from pyXLMS.exporter import to_pyxlinkviewer
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/pyxlinkviewer/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    pyxlinkviewer_result = to_pyxlinkviewer(
        crosslinks, pdb_file="6YHU", filename_prefix="6YHU"
    )
    assert pyxlinkviewer_result is not None
    pyxlinkviewer_output_file_str = pyxlinkviewer_result["PyXlinkViewer"]
    assert pyxlinkviewer_output_file_str is not None
    pyxlinkviewer_dataframe = pyxlinkviewer_result["PyXlinkViewer DataFrame"]
    assert pyxlinkviewer_dataframe is not None
    nr_mapped_crosslinks = pyxlinkviewer_result["Number of mapped crosslinks"]
    assert nr_mapped_crosslinks == 28
    crosslink_mapping = pyxlinkviewer_result["Mapping"]
    assert crosslink_mapping is not None
    parsed_pdb_sequenece = pyxlinkviewer_result["Parsed PDB sequence"]
    assert parsed_pdb_sequenece is not None
    parsed_pdb_chains = pyxlinkviewer_result["Parsed PDB chains"]
    assert parsed_pdb_chains is not None
    parsed_pdb_residue_numbers = pyxlinkviewer_result["Parsed PDB residue numbers"]
    assert parsed_pdb_residue_numbers is not None
    exported_files = pyxlinkviewer_result["Exported files"]
    assert len(exported_files) == 4
    for exported_file in exported_files:
        assert os.path.isfile(exported_file)


def test2():
    from pyXLMS.exporter import to_pyxlinkviewer
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/pyxlinkviewer/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    pyxlinkviewer_result = to_pyxlinkviewer(crosslinks, pdb_file="6YHU")
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test3():
    from pyXLMS.exporter import to_pyxlinkviewer
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/pyxlinkviewer/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    pyxlinkviewer_result = to_pyxlinkviewer(crosslinks, pdb_file=YHU_PDB)
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test4():
    from pyXLMS.exporter import to_pyxlinkviewer

    xls = helper_get_crosslinks(CROSSLINKS_ALL)
    pyxlinkviewer_result = to_pyxlinkviewer(xls, pdb_file=YHU_PDB)
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test5():
    from pyXLMS.exporter import to_pyxlinkviewer

    xls = helper_get_crosslinks(CROSSLINKS_ALL)
    pyxlinkviewer_result = to_pyxlinkviewer(xls, pdb_file="6YHU.pdb")
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test6():
    from pyXLMS.exporter import to_pyxlinkviewer

    xls = helper_get_crosslinks(CROSSLINKS_ALL)
    pyxlinkviewer_result = to_pyxlinkviewer(xls, pdb_file="6YHU")
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test7():
    from pyXLMS.exporter import to_pyxlinkviewer

    xls = helper_get_crosslinks(CROSSLINKS_ALL)
    pyxlinkviewer_result = to_pyxlinkviewer(xls, pdb_file="6yhu")
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test8():
    from pyXLMS.exporter import to_pyxlinkviewer

    xls = helper_get_crosslinks(CROSSLINKS_NSP8)
    pyxlinkviewer_result = to_pyxlinkviewer(xls, pdb_file="6yhu")
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_NSP8)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test9():
    from pyXLMS.exporter import to_pyxlinkviewer

    xls = helper_get_crosslinks(CROSSLINKS_NSP8)
    pyxlinkviewer_result = to_pyxlinkviewer(xls, pdb_file=JLT_PDB)
    should_result = helper_get_dataframe_from_pyxlinkviewer(JLT_RESULT_NSP8)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test10():
    from pyXLMS.exporter import to_pyxlinkviewer

    xls = helper_get_crosslinks(CROSSLINKS_NSP8)
    pyxlinkviewer_result = to_pyxlinkviewer(xls, pdb_file="7jlt")
    should_result = helper_get_dataframe_from_pyxlinkviewer(JLT_RESULT_NSP8)
    assert should_result.equals(pyxlinkviewer_result["PyXlinkViewer DataFrame"])


def test11():
    from pyXLMS.exporter import to_pyxlinkviewer
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/pyxlinkviewer/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    with pytest.raises(
        ValueError,
        match=r"Minimum sequence identity should be given as a fraction, e\.g\. 0\.8 for 80\% minimum sequence identity!",
    ):
        _pyxlinkviewer_result = to_pyxlinkviewer(
            crosslinks, pdb_file="6YHU", min_sequence_identity=80.0
        )


def test12():
    from pyXLMS.exporter import to_pyxlinkviewer

    with pytest.raises(ValueError, match=r"Provided crosslinks contain no elements!"):
        _pyxlinkviewer_result = to_pyxlinkviewer([], pdb_file="6YHU")


def test13():
    from pyXLMS.exporter import to_pyxlinkviewer
    from pyXLMS.data import create_csm_min

    csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    csms = [csm1]
    with pytest.raises(
        TypeError,
        match="Unsupported data type for input crosslinks! Parameter crosslinks has to be a list of crosslinks!",
    ):
        _pyxlinkviewer_result = to_pyxlinkviewer(csms, pdb_file="6YHU")


def test14():
    from pyXLMS.exporter import to_pyxlinkviewer
    from pyXLMS.data import create_csm_min
    from pyXLMS.data import create_crosslink_min

    csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    xl1 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    mixed = [xl1, csm1]
    with pytest.raises(
        TypeError, match="Not all elements in data have the same data type!"
    ):
        _pyxlinkviewer_result = to_pyxlinkviewer(mixed, pdb_file="6YHU")



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    xml = to_proxl(
        pr["crosslink-spectrum-matches"],
        fasta_filename="data/_fasta/Cas9_plus10.fasta",
        search_engine="MS Annika",
        search_engine_version="3.0.1",
        score="higher_better",
        crosslinker="DSS",
        filename="DSS_Cas9_ProXL.xml",
    )
    assert xml is not None


def test2():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl
    from lxml import etree

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    _xml = to_proxl(
        pr["crosslink-spectrum-matches"],
        fasta_filename="data/_fasta/Cas9_plus10.fasta",
        search_engine="MS Annika",
        search_engine_version="3.0.1",
        score="higher_better",
        crosslinker="DSS",
        filename="online.xml",
        schema_validation="online",
    )
    xmlschema_doc = etree.parse("data/_test/exporter/proxl/proxl-xml.xsd")
    xmlschema = etree.XMLSchema(xmlschema_doc)
    doc = etree.parse("online.xml")
    assert xmlschema.validate(doc)


def test3():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl
    from lxml import etree

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    _xml = to_proxl(
        pr["crosslink-spectrum-matches"],
        fasta_filename="data/_fasta/Cas9_plus10.fasta",
        search_engine="MS Annika",
        search_engine_version="3.0.1",
        score="higher_better",
        crosslinker="DSS",
        filename="offline.xml",
        schema_validation="offline",
    )
    xmlschema_doc = etree.parse("data/_test/exporter/proxl/proxl-xml.xsd")
    xmlschema = etree.XMLSchema(xmlschema_doc)
    doc = etree.parse("offline.xml")
    assert xmlschema.validate(doc)


def test4():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl
    from pyXLMS.transform import fasta_title_to_accession
    from lxml import etree

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    xml = to_proxl(
        pr["crosslink-spectrum-matches"],
        fasta_filename="data/_fasta/Cas9_plus10.fasta",
        search_engine="MS Annika",
        search_engine_version="3.0.1",
        score="higher_better",
        crosslinker="DSS",
        fasta_filename_override="gTuSC-parsimonious-plusRev.fasta",
        fasta_title_to_accession=fasta_title_to_accession,
        schema_validation="offline",
    )
    xmlschema_doc = etree.parse("data/_test/exporter/proxl/proxl-xml.xsd")
    xmlschema = etree.XMLSchema(xmlschema_doc)
    doc = etree.parse("offline.xml")
    assert xmlschema.validate(doc)
    assert 'fasta_filename="gTuSC-parsimonious-plusRev.fasta"' in xml
    assert '"sp|RETBP_HUMAN|"' not in xml


def test5():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    with pytest.raises(
        TypeError,
        match="Parameter 'score' has to be one of 'higher_better' or 'lower_better'!",
    ):
        _xml = to_proxl(
            pr["crosslink-spectrum-matches"],
            fasta_filename="data/_fasta/Cas9_plus10.fasta",
            search_engine="MS Annika",
            search_engine_version="3.0.1",
            score="greater_better",
            crosslinker="DSS",
            schema_validation="offline",
        )


def test6():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    with pytest.raises(
        TypeError,
        match="Parameter 'schema_validation' has to be one of 'online' or 'offline'!",
    ):
        _xml = to_proxl(
            pr["crosslink-spectrum-matches"],
            fasta_filename="data/_fasta/Cas9_plus10.fasta",
            search_engine="MS Annika",
            search_engine_version="3.0.1",
            score="higher_better",
            crosslinker="DSS",
            schema_validation="local",
        )


def test7():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    with pytest.raises(
        KeyError,
        match=(
            "Cannot infer crosslinker mass because crosslinker is not defined in "
            r"parameter 'modifications'\. Please specify crosslinker mass manually!"
        ),
    ):
        _xml = to_proxl(
            pr["crosslink-spectrum-matches"],
            fasta_filename="data/_fasta/Cas9_plus10.fasta",
            search_engine="MS Annika",
            search_engine_version="3.0.1",
            score="higher_better",
            crosslinker="SDA",
            schema_validation="offline",
        )


def test8():
    from pyXLMS.exporter import to_proxl

    with pytest.raises(
        ValueError,
        match="Provided crosslink-spectrum-matches contain no elements!",
    ):
        _xml = to_proxl(
            [],
            fasta_filename="data/_fasta/Cas9_plus10.fasta",
            search_engine="MS Annika",
            search_engine_version="3.0.1",
            score="higher_better",
            crosslinker="DSS",
            schema_validation="offline",
        )


def test9():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    with pytest.raises(
        TypeError,
        match="Unsupported data type for input csms! Parameter csms has to be a list of crosslink-spectrum-matches!",
    ):
        _xml = to_proxl(
            pr["crosslinks"],
            fasta_filename="data/_fasta/Cas9_plus10.fasta",
            search_engine="MS Annika",
            search_engine_version="3.0.1",
            score="higher_better",
            crosslinker="DSS",
            schema_validation="offline",
        )


def test10():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_proxl
    from pyXLMS.data import create_csm_min

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    with pytest.raises(
        RuntimeError,
        match="Can't export to ProXL because not all necessary information is available!",
    ):
        _xml = to_proxl(
            pr["crosslink-spectrum-matches"]
            + [create_csm_min("PEPK", 4, "KPEP", 1, "MSFile", 1)],
            fasta_filename="data/_fasta/Cas9_plus10.fasta",
            search_engine="MS Annika",
            search_engine_version="3.0.1",
            score="higher_better",
            crosslinker="DSS",
            schema_validation="offline",
        )


def test11():
    from pyXLMS import parser
    from pyXLMS import exporter
    from pyXLMS.transform import targets_only

    parser_result = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    csms = parser_result["crosslink-spectrum-matches"]
    csms = targets_only(csms)
    xml = exporter.to_proxl(
        csms,
        fasta_filename="data/_fasta/Cas9_plus10.fasta",
        search_engine="MS Annika",
        search_engine_version="3.0.1",
        score="higher_better",
        crosslinker="DSS",
        filename="CSMS_exported_to_ProXL.xml",
        schema_validation="online",
    )
    assert xml is not None



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest

XINET_COLS = [
    "Protein1",
    "PepPos1",
    "PepSeq1",
    "LinkPos1",
    "Protein2",
    "PepPos2",
    "PepSeq2",
    "LinkPos2",
    "Id",
]
XIVIEW_COLS = [
    "AbsPos1",
    "AbsPos2",
    "Protein1",
    "Protein2",
]


def test1():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslinks = targets_only(pr)["crosslinks"]
    cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    df = to_xiview(cas9, filename="crosslinks_xiVIEW.csv")
    assert df.shape[0] == 253
    assert df.shape[1] == 7


def test2():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslinks = targets_only(pr)["crosslinks"]
    cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    df = to_xiview(cas9, filename=None)
    assert df.shape[0] == 253
    assert df.shape[1] == 7


def test3():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslinks = targets_only(pr)["crosslinks"]
    cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    df = to_xiview(cas9, filename="crosslinks_xiVIEW.csv", minimal=False)
    assert df.shape[0] == 253
    assert df.shape[1] == 10


def test4():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslinks = targets_only(pr)["crosslinks"]
    cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    df = to_xiview(cas9, filename=None)
    assert df.shape[1] == 7
    cols = df.columns.values.tolist()
    for col in XIVIEW_COLS:
        assert col in cols
    assert "Decoy1" in cols
    assert "Decoy2" in cols
    assert "Score" in cols


def test5():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslinks = targets_only(pr)["crosslinks"]
    cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    for xl in cas9:
        xl["score"] = None
        xl["alpha_decoy"] = None
    df = to_xiview(cas9, filename=None)
    assert df.shape[1] == 4
    cols = df.columns.values.tolist()
    for col in XIVIEW_COLS:
        assert col in cols
    assert "Decoy1" not in cols
    assert "Decoy2" not in cols
    assert "Score" not in cols


def test6():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslinks = targets_only(pr)["crosslinks"]
    cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    df = to_xiview(cas9, filename=None, minimal=False)
    assert df.shape[1] == 10
    cols = df.columns.values.tolist()
    for col in XINET_COLS:
        assert col in cols
    assert "Score" in cols


def test7():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslinks = targets_only(pr)["crosslinks"]
    cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    for xl in cas9:
        xl["score"] = None
    df = to_xiview(cas9, filename=None, minimal=False)
    assert df.shape[1] == 9
    cols = df.columns.values.tolist()
    for col in XINET_COLS:
        assert col in cols
    assert "Score" not in cols


def test8():
    from pyXLMS.exporter import to_xiview

    with pytest.raises(
        ValueError,
        match="Provided crosslinks contain no elements!",
    ):
        _df = to_xiview([], filename="crosslinks_xiVIEW.csv")


def test9():
    from pyXLMS.exporter import to_xiview
    from pyXLMS.data import create_crosslink_min

    xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    crosslinks = [xl1, xl2]

    with pytest.raises(
        RuntimeError,
        match="Can't export to xiVIEW because not all necessary information is available!",
    ):
        _df = to_xiview(crosslinks, filename=None)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import warnings
from tqdm import tqdm

from ..data import check_input
from ..data import check_input_multi
from ..data import create_parser_result
from .util import get_available_keys
from .filter import filter_target_decoy
from .filter import filter_crosslink_type

from typing import Dict
from typing import List
from typing import Any

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def __annotate_fdr_strict(
    data: List[Dict[str, Any]],
    score: Literal["higher_better", "lower_better"],
) -> List[Dict[str, Any]]:
    r"""Annotates a list of crosslinks or crosslink-spectrum-matches with their false discover rate by strict false discovery rate estimation.

    Annotates a list of crosslinks or crosslink-spectrum-matches with their false discover rate (FDR) by strict false discovery rate estimation
    using the formula (TD+DD)/TT. The annotated FDR is written to "additional_information" for every crosslink or crosslink-spectrum-match and
    is available via the accessor "pyXLMS_annotated_FDR".
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to annotate.
    score: str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any
        A list of FDR annotated crosslink-spectrum-matches or crosslinks.

    Warns
    -----
    RuntimeWarning
        If one of the false discovery rates can't be computed because of not enough target-target matches.

    Notes
    -----
    This function should not be called directly, it is called from ``annotate_fdr()``.
    """
    data_type = (
        "crosslinks"
        if data[0]["data_type"] == "crosslink"
        else "crosslink-spectrum-matches"
    )
    for current_item in tqdm(
        data,
        total=len(data),
        desc=f"Annotating FDR for {data_type}...",
    ):
        tt = 0
        td = 0
        dd = 0
        for item in data:
            if score == "higher_better":
                if item["score"] >= current_item["score"]:
                    if item["alpha_decoy"] and item["beta_decoy"]:
                        dd += 1
                    elif not item["alpha_decoy"] and not item["beta_decoy"]:
                        tt += 1
                    else:
                        td += 1
            else:
                if item["score"] <= current_item["score"]:
                    if item["alpha_decoy"] and item["beta_decoy"]:
                        dd += 1
                    elif not item["alpha_decoy"] and not item["beta_decoy"]:
                        tt += 1
                    else:
                        td += 1
        fdr = float("nan")
        if tt == 0:
            warnings.warn(
                RuntimeWarning(
                    f"Could not calculate FDR {data[0]['data_type']} because the number of target-target matches "
                    + f"at score {current_item['score']} is 0!"
                )
            )
        else:
            fdr = (td + dd) / tt
        if current_item["additional_information"] is None:
            current_item["additional_information"] = {"pyXLMS_annotated_FDR": fdr}
        else:
            current_item["additional_information"]["pyXLMS_annotated_FDR"] = fdr
    return data


def __annotate_fdr_relaxed(
    data: List[Dict[str, Any]],
    score: Literal["higher_better", "lower_better"],
) -> List[Dict[str, Any]]:
    r"""Annotates a list of crosslinks or crosslink-spectrum-matches with their false discover rate by relaxed false discovery rate estimation.

    Annotates a list of crosslinks or crosslink-spectrum-matches with their false discover rate (FDR) by relaxed false discovery rate estimation
    using the formula (TD-DD)/TT. The annotated FDR is written to "additional_information" for every crosslink or crosslink-spectrum-match and
    is available via the accessor "pyXLMS_annotated_FDR".
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to annotate.
    score: str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any
        A list of FDR annotated crosslink-spectrum-matches or crosslinks.

    Warns
    -----
    RuntimeWarning
        If one of the false discovery rates can't be computed because of not enough target-target matches.
    RuntimeWarning
        If one of the false discovery rates can't be computed because the number of decoy-decoy matches exceeds
        the number of target-decoy matches.

    Notes
    -----
    This function should not be called directly, it is called from ``annotate_fdr()``.
    """
    data_type = (
        "crosslinks"
        if data[0]["data_type"] == "crosslink"
        else "crosslink-spectrum-matches"
    )
    for current_item in tqdm(
        data,
        total=len(data),
        desc=f"Annotating FDR for {data_type}...",
    ):
        tt = 0
        td = 0
        dd = 0
        for item in data:
            if score == "higher_better":
                if item["score"] >= current_item["score"]:
                    if item["alpha_decoy"] and item["beta_decoy"]:
                        dd += 1
                    elif not item["alpha_decoy"] and not item["beta_decoy"]:
                        tt += 1
                    else:
                        td += 1
            else:
                if item["score"] <= current_item["score"]:
                    if item["alpha_decoy"] and item["beta_decoy"]:
                        dd += 1
                    elif not item["alpha_decoy"] and not item["beta_decoy"]:
                        tt += 1
                    else:
                        td += 1
        fdr = float("nan")
        if tt == 0:
            warnings.warn(
                RuntimeWarning(
                    f"Could not calculate FDR {data[0]['data_type']} because the number of target-target matches "
                    + f"at score {current_item['score']} is 0!"
                )
            )
        elif td - dd < 0:
            warnings.warn(
                RuntimeWarning(
                    f"Could not calculate FDR {data[0]['data_type']} because number the of decoy-decoy matches "
                    + f"exceeds the number of target-decoy matches at score {current_item['score']}!"
                )
            )
        else:
            fdr = (td + dd) / tt
        if current_item["additional_information"] is None:
            current_item["additional_information"] = {"pyXLMS_annotated_FDR": fdr}
        else:
            current_item["additional_information"]["pyXLMS_annotated_FDR"] = fdr
    return data


def annotate_fdr(
    data: List[Dict[str, Any]] | Dict[str, Any],
    formula: Literal["D/T", "(TD+DD)/TT", "(TD-DD)/TT"] = "D/T",
    score: Literal["higher_better", "lower_better"] = "higher_better",
    separate_intra_inter: bool = False,
    ignore_missing_labels: bool = False,
) -> List[Dict[str, Any]] | Dict[str, Any]:
    r"""Annotates a list of crosslinks or crosslink-spectrum-matches, or a parser_result with their false dicovery rate by estimating the false discovery rate.

    Annotates a list of crosslinks or crosslink-spectrum-matches, or a parser_result with their false discovery rate (FDR) by estimating false discovery rate
    using the defined formula. The annotated FDR is written to "additional_information" for every crosslink or crosslink-spectrum-match and is available via
    the accessor "pyXLMS_annotated_FDR".
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any, or dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to annotate, or a parser_result.
    formula : str, one of "D/T", "(TD+DD)/TT", or "(TD-DD)/TT", default = "D/T"
        Which formula to use to estimate FDR. D and DD denote decoy matches, T and TT denote target matches, and TD denotes target-decoy
        and decoy-target matches.
    score: str, one of "higher_better" or "lower_better", default = "higher_better"
        If a higher score is considered better, or a lower score is considered better.
    separate_intra_inter : bool, default = False
        If FDR should be estimated separately for intra and inter matches.
    ignore_missing_labels : bool, default = False
        If crosslinks and crosslink-spectrum-matches should be ignored if they don't have target and decoy labels. By default and error is
        thrown if any unlabelled data is encountered.

    Returns
    -------
    list of dict of str, any, or dict of str, any
        If a list of crosslink-spectrum-matches or crosslinks was provided, a list of annotated
        crosslink-spectrum-matches or crosslinks is returned. If a parser_result was provided,
        an parser_result with annotated crosslink-spectrum-matches and/or annotated crosslinks will
        be returned.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If parameter formula is not one of 'D/T', '(TD+DD)/TT', or '(TD-DD)/TT'.
    TypeError
        If parameter score is not one of 'higher_better' or 'lower_better'.
    ValueError
        If attribute 'score' is not available for any of the data.
    ValueError
        If attribute 'alpha_decoy' or 'beta_decoy' is not available for any of the data and parameter ignore_missing_labels
        is set to False.
    ValueError
        If the number of DD matches exceeds the number of TD matches for formula '(TD-DD)/TT'.
        FDR cannot be annotated with the formula '(TD-DD)/TT' in these cases.

    Examples
    --------
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import annotate_fdr
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> csms = pr["crosslink-spectrum-matches"]
    >>> csms = annotate_fdr(csms)
    >>> validated_csms = [
    ...     csm
    ...     for csm in csms
    ...     if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ... ]
    >>> len(validated_csms)
    705

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import annotate_fdr
    >>> pr = read(
    ...     [
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     ],
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> pr = annotate_fdr(pr)
    >>> validated_csms = [
    ...     csm
    ...     for csm in pr["crosslink-spectrum-matches"]
    ...     if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ... ]
    >>> len(validated_csms)
    705
    >>> validated_xls = [
    ...     xl
    ...     for xl in pr["crosslinks"]
    ...     if xl["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ... ]
    >>> len(validated_xls)
    226
    """
    _ok = check_input_multi(data, "data", [dict, list])
    _ok = check_input(formula, "formula", str)
    _ok = check_input(score, "score", str)
    _ok = check_input(separate_intra_inter, "separate_intra_inter", bool)
    _ok = check_input(ignore_missing_labels, "ignore_missing_labels", bool)
    if formula not in ["D/T", "(TD+DD)/TT", "(TD-DD)/TT"]:
        raise TypeError(
            "Parameter 'formula' has to be one of 'D/T', '(TD+DD)/TT' or '(TD-DD)/TT'! Where D and DD is the number of decoys, T and TT the number of targets, "
            "and TD the number of target-decoys!"
        )
    if score not in ["higher_better", "lower_better"]:
        raise TypeError(
            "Parameter 'score' has to be one of 'higher_better' or 'lower_better'!"
        )
    if isinstance(data, list):
        _ok = check_input(data, "data", list, dict)
        if len(data) == 0:
            return data
        if "data_type" not in data[0] or data[0]["data_type"] not in [
            "crosslink",
            "crosslink-spectrum-match",
        ]:
            raise TypeError(
                "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match, "
                "or a parser_result!"
            )
        if ignore_missing_labels:
            data = [
                item
                for item in data
                if item["alpha_decoy"] is not None and item["beta_decoy"] is not None
            ]
        available_keys = get_available_keys(data)
        if (
            not available_keys["score"]
            or not available_keys["alpha_decoy"]
            or not available_keys["beta_decoy"]
        ):
            raise ValueError(
                "Can't annotate data if 'score' or target/decoy labels are missing! Selecting 'ignore_missing_labels = True' will ignore crosslinks and crosslink-spectrum-matches "
                "that don't have a valid target/decoy label and filter them out!"
            )
        if formula == "(TD-DD)/TT":
            if len(filter_target_decoy(data)["Target-Decoy"]) == 0:
                raise ValueError(
                    "Can't annotate FDR with formula '(TD-DD)/TT' when there are no TD matches! Please select the default formula instead!"
                )
            if separate_intra_inter:
                separate = filter_crosslink_type(data)
                return __annotate_fdr_relaxed(
                    separate["Intra"], score
                ) + __annotate_fdr_relaxed(separate["Inter"], score)
            return __annotate_fdr_relaxed(data, score)
        if separate_intra_inter:
            separate = filter_crosslink_type(data)
            return __annotate_fdr_strict(
                separate["Intra"], score
            ) + __annotate_fdr_strict(separate["Inter"], score)
        return __annotate_fdr_strict(data, score)
    if "data_type" not in data or data["data_type"] != "parser_result":
        raise TypeError("Can't validate dict. Dict has to be a valid 'parser_result'!")
    new_csms = (
        annotate_fdr(
            data["crosslink-spectrum-matches"],
            formula,
            score,
            separate_intra_inter,
            ignore_missing_labels,
        )
        if data["crosslink-spectrum-matches"] is not None
        else None
    )
    new_xls = (
        annotate_fdr(
            data["crosslinks"],
            formula,
            score,
            separate_intra_inter,
            ignore_missing_labels,
        )
        if data["crosslinks"] is not None
        else None
    )
    if new_csms is not None:
        if not isinstance(new_csms, list):
            raise RuntimeError(
                "Something went wrong while annotating crosslink-spectrum-matches.\n"
                f"Expected data type: list. Got: {type(new_csms)}."
            )
    if new_xls is not None:
        if not isinstance(new_xls, list):
            raise RuntimeError(
                "Something went wrong while annotating crosslinks.\n"
                f"Expected data type: list. Got: {type(new_xls)}."
            )
    return create_parser_result(
        search_engine=data["search_engine"], csms=new_csms, crosslinks=new_xls
    )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest

FILES = {
    "data/maxquant/run1/crosslinkMsms.txt": {
        "engine": "MaxQuant",
        "crosslinker": "DSS",
    },
    "data/merox/XLpeplib_Beveridge_QEx-HFX_DSS_R1.csv": {
        "engine": "MeroX",
        "crosslinker": "DSS",
    },
    "data/merox/XLpeplib_Beveridge_QEx-HFX_DSS_R1.zhrm": {
        "engine": "MeroX",
        "crosslinker": "DSS",
    },
    "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.mzid": {
        "engine": "mzIdentML",
        "crosslinker": "DSS",
    },
    "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult": {
        "engine": "MS Annika",
        "crosslinker": "DSS",
    },
    "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.txt": {
        "engine": "MS Annika",
        "crosslinker": "DSS",
    },
    "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx": {
        "engine": "MS Annika",
        "crosslinker": "DSS",
    },
    "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt": {
        "engine": "MS Annika",
        "crosslinker": "DSS",
    },
    "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx": {
        "engine": "MS Annika",
        "crosslinker": "DSS",
    },
    "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv": {
        "engine": "pLink",
        "crosslinker": "DSS",
    },
    "data/plink3/Cas10_plus10_2025.04.07.filtered_cross-linked_spectra.csv": {
        "engine": "pLink",
        "crosslinker": "DSS",
    },
    "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_peptides.csv": {
        "engine": "pLink",
        "crosslinker": "DSS",
    },
    "data/plink3/Cas10_plus10_2025.04.07.filtered_cross-linked_peptides.csv": {
        "engine": "pLink",
        "crosslinker": "DSS",
    },
    "data/pyxlms/csm.txt": {"engine": "Custom", "crosslinker": "DSS"},
    "data/pyxlms/xl.txt": {"engine": "Custom", "crosslinker": "DSS"},
    "data/scout/Cas9_Filtered_CSMs.csv": {"engine": "Scout", "crosslinker": "DSSO"},
    "data/scout/Cas9_Residue_Pairs.csv": {"engine": "Scout", "crosslinker": "DSSO"},
    "data/scout/Cas9_Unfiltered_CSMs.csv": {"engine": "Scout", "crosslinker": "DSSO"},
    "data/xi/1perc_xl_boost_CSM_xiFDR2.2.1.csv": {
        "engine": "xiSearch/xiFDR",
        "crosslinker": "BS3",
    },
    "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv": {
        "engine": "xiSearch/xiFDR",
        "crosslinker": "BS3",
    },
    "data/xi/r1_Xi1.7.6.7.csv": {"engine": "xiSearch/xiFDR", "crosslinker": "BS3"},
    "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3.mzid": {
        "engine": "mzIdentML",
        "crosslinker": "DSSO",
    },
    "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3.pdResult": {
        "engine": "XlinkX",
        "crosslinker": "DSSO",
    },
    "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3_Crosslinks.txt": {
        "engine": "XlinkX",
        "crosslinker": "DSSO",
    },
    "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3_Crosslinks.xlsx": {
        "engine": "XlinkX",
        "crosslinker": "DSSO",
    },
    "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3_CSMs.txt": {
        "engine": "XlinkX",
        "crosslinker": "DSSO",
    },
    "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3_CSMs.xlsx": {
        "engine": "XlinkX",
        "crosslinker": "DSSO",
    },
}


def test1():
    from pyXLMS import parser as p

    with pytest.raises(ValueError):
        _r = p.read(
            "data/maxquant/run1/crosslinkMsms.txt",
            engine="MS Amanda",
            crosslinker="DSS",
        )


@pytest.mark.slow
def test2():
    from pyXLMS import parser as p

    for k, v in FILES.items():
        pr = p.read(k, engine=v["engine"], crosslinker=v["crosslinker"], verbose=0)
        assert pr["search_engine"] == v["engine"]
